"""
/**
 * @file 计算结果文件的均值
 * @author weiqiangdi@gmail.com
 */
"""
import os
from os.path import join
import glob
import numpy as np


def cal_means(all_res_files, data_name, filter_outlier=False):
    recall_res = []
    ndcg_res = []
    valid_file_num = 0
    invalid_file_index = []

    if all_res_files:
        for res_file in all_res_files:
            with open(res_file, 'r') as log_file:
                valid_flag = False
                file_index = os.path.splitext(os.path.basename(res_file))[0].split('_')[-1]
                for line in log_file:
                    if 'best recall' in line:
                        # 主要是为了防止对运行途中中断导致没有最终结果的文件的错误计数，所以没有采用len(all_res_files)的方式
                        valid_file_num = valid_file_num + 1
                        valid_flag = True
                        # https://stackoverflow.com/questions/41535638/extract-and-get-value-from-a-text-file-python
                        recall_res.append(float(line.split(",")[1].split(":")[1].strip()))

                    if 'best ndcg' in line:
                        # https://stackoverflow.com/questions/41535638/extract-and-get-value-from-a-text-file-python
                        ndcg_res.append(float(line.split(",")[2].split(":")[1].strip()))
                if not valid_flag:
                    # 记录下无效文件的索引
                    invalid_file_index.append(file_index)

        if filter_outlier:
            # 去掉最g高异常值, 注意当该步文件只有一个时则会警告
            min_ind = recall_res.index(min(recall_res))
            recall_res.pop(min_ind)
            ndcg_res.pop(min_ind)

        # https://stackoverflow.com/questions/29217399/convert-list-of-numpy-float64-to-float-in-python-quickly
        mean_recall = round(float(np.mean(recall_res)), 4)
        mean_ndcg = round(float(np.mean(ndcg_res)), 4)

        print(f'filter_outlier: {filter_outlier}')
        print(f'{data_name}_recall_means: {mean_recall}\t {data_name}_ndcg_means: {mean_ndcg}'
              f'\t valid files nums: {valid_file_num}\t invalid file index: {invalid_file_index}')
    else:
        print(f'do not have result files of {data_name}!')


log_dir = './log_dir'
# https://stackoverflow.com/questions/11968976/list-files-only-in-the-current-directory
# files = [log_file for log_file in os.listdir(log_dir) if os.path.isfile(join(log_dir, log_file))]

# https://stackoverflow.com/questions/3964681/find-all-files-in-a-directory-with-extension-txt-in-python
cds_res_files = glob.glob(join(log_dir, 'cds_res_*.txt'))
comics_res_files = glob.glob(join(log_dir, 'comics_res_*.txt'))

cal_means(cds_res_files, 'cds')
cal_means(comics_res_files, 'comics')
print('************************************************************************')
cal_means(cds_res_files, 'cds', filter_outlier=True)
cal_means(comics_res_files, 'comics', filter_outlier=True)
