import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable

device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')


class HGN(nn.Module):
    # 17052    35119
    def __init__(self, num_users, num_items, args, device):
        super(HGN, self).__init__()

        self.args = args

        # init args
        # 5
        L = self.args.L
        # 50
        d_model = args.d_model

        # user and item embeddings
        self.user_embeddings = nn.Embedding(num_users, d_model).to(device)
        self.item_embeddings = nn.Embedding(num_items, d_model).to(device)

        self.W2 = nn.Embedding(num_items, d_model, padding_idx=0).to(device)
        self.b2 = nn.Embedding(num_items, 1, padding_idx=0).to(device)

        # weight initialization
        self.user_embeddings.weight.data.normal_(0, 1.0 / self.user_embeddings.embedding_dim)
        self.item_embeddings.weight.data.normal_(0, 1.0 / self.item_embeddings.embedding_dim)
        self.W2.weight.data.normal_(0, 1.0 / self.W2.embedding_dim)
        self.b2.weight.data.zero_()

        # self.in_space_att = SpaceAttBlock(args)
        # self.conv_part = ConvBlock(args.L, args)

        if args.has_drop:
            self.drop_item = nn.Dropout(0.5)
            self.drop_user = nn.Dropout(0.5)

    # torch.Size([4096, 5])    torch.Size([4096])   torch.Size([4096, 6])
    # 当测试时 batch size: 1024 且 items_to_predict: torch.Size([35119])
    def forward(self, item_seq, user_ids, items_to_predict, for_pred=False):
        # torch.Size([4096, 5, 50])
        item_embs = self.item_embeddings(item_seq)
        # torch.Size([4096, 50])
        user_emb = self.user_embeddings(user_ids)
        if self.args.has_drop:
            item_embs = self.drop_item(item_embs)
            user_emb = self.drop_user(user_emb)

        # torch.Size([4096, 5, 50])
        # in_att_out = self.in_space_att(item_embs, user_emb)
        # torch.Size([4096, 5, 50])
        # conv_out = self.conv_part(item_embs)

        # todo 下面sum可能不好，可以试试取最后一行
        # torch.Size([4096, 50])
        # union_out = item_embs[:, -1, :]
        # 相当于求均值，但不是传统意义下的均值，比如若instance_score中系数全为1，就退化为传统系数下的均值，若全部为0.5，
        # 那么最后除以的就是2.5而非5，应该可以理解为和上一步系数耦合的加权平均
        # torch.Size([4096, 50])
        # todo 下面这步其实是可以融入到in_space_att里面的，就是可以提前除
        # union_out = union_out / torch.sum(instance_score, dim=1).unsqueeze(1)

        # torch.Size([4096, 6, 50])    测试时: torch.Size([35119, 50])
        w2 = self.W2(items_to_predict)
        # torch.Size([4096, 6, 1])     测试时: torch.Size([35119, 1])
        b2 = self.b2(items_to_predict)

        len = item_embs.size(1)
        raw_posi = torch.zeros(len).to(device)
        for i in range(1, len + 1):
            decayed_val = 0.3 ** (len - i)
            raw_posi[i - 1] = decayed_val

        if for_pred:
            w2 = w2.squeeze()
            # torch.Size([35119])
            b2 = b2.squeeze()

            # torch.Size([1024, 35119, 50])
            user_emb_dyn = torch.bmm(torch.bmm(w2.unsqueeze(0).expand(item_embs.size(0), -1, -1), item_embs.permute(0, 2, 1)).softmax(-1), item_embs)

            # user_emb_dyn =

            # MF
            # torch.Size([1024, 35119]) 注意测试时batch改变为1024了
            res = user_emb.mm(w2.t()) + b2
            res += (w2.unsqueeze(0) * user_emb_dyn).sum(2)

            # union-level
            # res += union_out.mm(w2.t())

            # item-item product
            # torch.Size([1024, 5, 35119])
            rel_score = torch.matmul(item_embs, w2.t().unsqueeze(0))
            rel_score = rel_score * raw_posi.unsqueeze(1).expand(-1, rel_score.size(2)).unsqueeze(0).expand(rel_score.size(0), -1, -1)
            rel_score = torch.sum(rel_score, dim=1)
            res += rel_score
        else:
            # torch.Size([4096, 6, 50])
            user_emb_dyn = torch.bmm(torch.bmm(w2, item_embs.permute(0, 2, 1)).softmax(-1), item_embs)
            # MF
            # 这个计算的是式5中的第一项, out = input + batch1@batch2
            # torch.Size([4096, 6])
            res = torch.baddbmm(b2, w2, user_emb.unsqueeze(2)).squeeze()
            res += (w2 * user_emb_dyn).sum(2)

            # temp = torch.bmm(union_out.unsqueeze(1), w2.permute(0, 2, 1))  # torch.Size([4096, 1, 6])
            # union-level
            # 这个计算的是式5中的第二项
            # res += torch.bmm(union_out.unsqueeze(1), w2.permute(0, 2, 1)).squeeze()

            # todo item-item product （仅起强调作用）
            # 这个计算的是式5中的第三项
            # torch.Size([4096, 5, 6])
            rel_score = item_embs.bmm(w2.permute(0, 2, 1))
            rel_score = rel_score * raw_posi.unsqueeze(1).expand(-1, rel_score.size(2)).unsqueeze(0).expand(rel_score.size(0), -1, -1)
            rel_score = torch.sum(rel_score, dim=1)
            res += rel_score

        return res


class SpaceAttBlock(nn.Module):
    def __init__(self, args):
        super(SpaceAttBlock, self).__init__()

        d_model = args.d_model
        self.feature_gate_item = nn.Linear(d_model, d_model).to(args.device)
        self.feature_gate_user = nn.Linear(d_model, d_model).to(args.device)

        # 其实相当于一个线性层
        self.instance_gate_item = Variable(torch.zeros(d_model, 1).type(torch.FloatTensor), requires_grad=True).to(
            args.device)
        self.instance_gate_user = Variable(torch.zeros(d_model, args.L).type(torch.FloatTensor), requires_grad=True).to(
            args.device)
        self.instance_gate_item = torch.nn.init.xavier_uniform_(self.instance_gate_item)
        self.instance_gate_user = torch.nn.init.xavier_uniform_(self.instance_gate_user)

    # torch.Size([1024, 8, 168])
    def forward(self, item_embs, user_emb):
        # feature gating
        # torch.Size([4096, 5, 50])
        gate = torch.sigmoid(self.feature_gate_item(item_embs) + self.feature_gate_user(user_emb).unsqueeze(1))
        # torch.Size([4096, 5, 50])
        gated_item = item_embs * gate

        # instance gating
        # torch.Size([4096, 5])
        instance_score = torch.sigmoid(torch.matmul(gated_item, self.instance_gate_item.unsqueeze(0)).squeeze() +
                                       user_emb.mm(self.instance_gate_user))
        # torch.Size([4096, 5, 50])
        union_out = gated_item * instance_score.unsqueeze(2)
        union_out = union_out / torch.sum(instance_score, dim=1).unsqueeze(1).unsqueeze(1)

        return union_out


class ConvBlock(nn.Module):
    def __init__(self, in_channel, args):
        super(ConvBlock, self).__init__()

        self.conv_layers = nn.ModuleList(
            [ConvLayer(in_channel, 2 ** i, args) for i in range(args.conv_layer)]
        )

    def forward(self, input):
        conv_block_out = input
        for conv_op in self.conv_layers:
            conv_block_out = conv_op(conv_block_out)

        return conv_block_out


class ConvLayer(nn.Module):
    def __init__(self, in_channel, di_rate, args):
        super(ConvLayer, self).__init__()

        self.conv_blocks = nn.ModuleList(
            [MultibranchBlock(in_channel, di_rate, args) for _ in range(2)]
        )

    def forward(self, input):
        conv_layer_out = input
        for conv_op in self.conv_blocks:
            conv_layer_out = conv_op(conv_layer_out)

        return conv_layer_out + input


class MultibranchBlock(nn.Module):
    def __init__(self, in_channel, di_rate, args):
        super(MultibranchBlock, self).__init__()

        pool_size = 7

        self.conv_h = nn.Conv1d(in_channels=pool_size,
                                out_channels=pool_size,
                                kernel_size=3,
                                dilation=di_rate,
                                padding=(3 - 1) * di_rate // 2,
                                groups=pool_size)
        self.conv_c = nn.Conv1d(in_channels=3,
                                out_channels=3,
                                kernel_size=9,
                                dilation=di_rate,
                                padding=(9 - 1) * di_rate // 2,
                                groups=3)
        self.conv_3 = nn.Conv1d(in_channels=in_channel,
                                out_channels=in_channel,
                                kernel_size=3,
                                # dilation=di_rate,
                                padding=(3 - 1) * 1 // 2,
                                groups=in_channel)
        self.act = nn.ReLU()
        self.avg_adap_h = nn.AdaptiveAvgPool1d(output_size=pool_size)

    def forward(self, input):
        B, L, H = input.size()
        # torch.Size([4096, 7, 5])
        strip_h = self.avg_adap_h(input).permute(0, 2, 1)
        # torch.Size([4096, 3, 50])
        # strip_c = self.avg_adap_c(input.permute(0, 2, 1)).permute(0, 2, 1)
        # torch.Size([4096, 7, 5])
        conv_strip_h = self.conv_h(strip_h)

        interpolate_h_temp = [item.unsqueeze(2).expand(-1, L, H) for item in conv_strip_h.permute(1, 0, 2)]
        # interpolate_c_temp = [item.unsqueeze(1).expand(-1, L, H) for item in conv_strip_c.permute(1, 0, 2)]
        # torch.Size([4096, 5, 50])
        interpolate_h = torch.stack(interpolate_h_temp, 3).mean(-1)
        # torch.Size([4096, 5, 50])
        # interpolate_c = torch.stack(interpolate_c_temp, 3).mean(-1)
        acted_out = input * interpolate_h.sigmoid()
        # conv_strip_h = self.att_layer(conv_strip_h.mean(-1)).unsqueeze(2) * conv_strip_h
        # torch.Size([4096, 3, 50])
        # conv_strip_c = self.conv_c(strip_c)
        # conv_strip_c = self.att_layer2(conv_strip_c) * conv_strip_c
        # interpolate_h_temp = [item.unsqueeze(2).expand(-1, L, H) for item in conv_strip_h.permute(1, 0, 2)]
        # interpolate_c_temp = [item.unsqueeze(1).expand(-1, L, H) for item in conv_strip_c.permute(1, 0, 2)]
        # torch.Size([4096, 5, 50])
        # interpolate_h = conv_strip_h.mean(-2)
        # torch.Size([4096, 5, 50])
        # interpolate_c = torch.stack(interpolate_c_temp, 3).mean(-1)
        # acted_out = input * interpolate_h.sigmoid().unsqueeze(2)

        return acted_out
