# from model.gating_network import HGN
from interactions import Interactions
from eval_metrics import *
from models import *
from utils import *
import argparse
# import logging
from time import time
import datetime
import torch
import sys

from data import Amazon, MovieLens, GoodReads

device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')


# test_set: list(17052)
def evaluation(model, train, test_set, args, topk=20):
    # 17052
    num_users = train.num_users
    # 35119
    num_items = train.num_items
    batch_size = args.batch_eval
    # 17
    num_batches = int(num_users / batch_size) + 1
    user_indexes = np.arange(num_users)
    item_indexes = np.arange(num_items)
    pred_list = None
    # 注意到，train.tocsr()只是额外生成一个类似邻接矩阵的train_matrix，并不影响train里面原本存储的样本
    # <class 'tuple'>: (17052, 35119)
    train_matrix = train.tocsr()
    # <class 'tuple'>: (17052, 5)
    test_sequences = train.test_sequences.sequences

    for batchID in range(num_batches):
        start = batchID * batch_size
        end = start + batch_size

        if batchID == num_batches - 1:
            if start < num_users:
                end = num_users
            else:
                break

        # <class 'tuple'>: (1024,)
        batch_user_index = user_indexes[start:end]

        # <class 'tuple'>: (1024, 5)
        batch_test_sequences = test_sequences[batch_user_index]
        # 判断batch_test_sequences，如果小于2维，则变为2维数据，高维不影响
        batch_test_sequences = np.atleast_2d(batch_test_sequences)

        batch_test_sequences = torch.from_numpy(batch_test_sequences).type(torch.LongTensor).to(device)
        # torch.Size([35119])
        item_ids = torch.from_numpy(item_indexes).type(torch.LongTensor).to(device)
        # torch.Size([1024])
        batch_user_ids = torch.from_numpy(np.array(batch_user_index)).type(torch.LongTensor).to(device)

        # torch.Size([1024, 35119])
        rating_pred = model(batch_test_sequences, batch_user_ids, item_ids, True)
        rating_pred = rating_pred.cpu().data.numpy().copy()
        rating_pred[train_matrix[batch_user_index].toarray() > 0] = 0

        # reference: https://stackoverflow.com/a/23734295, https://stackoverflow.com/a/20104162
        # <class 'tuple'>: (1024, 35119)
        ind = np.argpartition(rating_pred, -topk)
        # <class 'tuple'>: (1024, 20)
        ind = ind[:, -topk:]
        # temp = len(rating_pred)  # 1024
        # 下面涉及到[np.arange(len(rating_pred))[:, None]的都用到了上面那行第二个链接的内容：https://stackoverflow.com/a/20104162
        # <class 'tuple'>: (1024, 20)
        arr_ind = rating_pred[np.arange(len(rating_pred))[:, None], ind]
        # arr_ind虽是最大的几个，但内部是无序的，可参考https://stackoverflow.com/a/23734295，所以想内部排序的话还需要np.argsort，
        # 后面的[np.arange(len(rating_pred)), ::-1]是反序, 反序排列是为了后面计算precision等指标时方便
        # <class 'tuple'>: (1024, 20)   可拿右边简单做个验证 a = np.array([[0,4,3], [5,6,0], [0,9,2], [8,7,6]])
        arr_ind_argsort = np.argsort(arr_ind)[np.arange(len(rating_pred)), ::-1]
        # 注意，我们最终需要的不是rating_pred中最大的那几个值，而是最大那几个值对应的索引，也就是item的序号
        # <class 'tuple'>: (1024, 20)
        batch_pred_list = ind[np.arange(len(rating_pred))[:, None], arr_ind_argsort]

        if batchID == 0:
            pred_list = batch_pred_list
        else:
            pred_list = np.append(pred_list, batch_pred_list, axis=0)

    recall, ndcg = [], []
    for k in [10, 20]:
        # precision.append(precision_at_k(test_set, pred_list, k))
        recall.append(recall_at_k(test_set, pred_list, k))
        # MAP.append(mapk(test_set, pred_list, k))
        ndcg.append(ndcg_k(test_set, pred_list, k))

    return recall, ndcg


# <class 'tuple'>: (26,)    35119   150
def negsamp_vectorized_bsearch_preverif(pos_inds, n_items, n_samp=32):
    """ Pre-verified with binary search
    `pos_inds` is assumed to be ordered
    reference: https://tech.hbc.com/2018-03-23-negative-sampling-in-numpy.html
    """
    # <class 'tuple'>: (150,)
    raw_samp = np.random.randint(0, n_items - len(pos_inds), size=n_samp)
    # <class 'tuple'>: (26,)
    pos_inds_adj = pos_inds - np.arange(len(pos_inds))
    # temp = np.searchsorted(pos_inds_adj, raw_samp, side='right')  # <class 'tuple'>: (150,)
    # 整体的思想其实并不难，就是在当前user没有连接的item里随机采样负样本，其实可参考Caser，那里应该有更容易理解的做法。
    # 这里我们分情况讨论，假设有1000个item，用户已经有连接的有50个，就是pos_inds，想采样100个负样本。先从剩下的950个序号里选。
    # 对于raw_samp里面的某个采样值，如果它小于pos_inds_adj的所有值，显然采集到的这个点不会是已经连接的那些点，通过反证法很容易证得，如果是，
    # 那么加上50必然等于已经连接的某个值，或者说采集到的这个点必然等于已经连接的点减去50后的某个值，即pos_inds_adj中的值，
    # 此时np.searchsorted中对应该点的值为0，raw_samp中对应的该点加上0还是该点，原因很简单，该采集的点是合格的，所以不用变
    # 所以说下面的np.searchsorted最关键的是等于的那种情况，必须要加1来避免，当raw_samp中的某点大于pos_inds_adj中的某点小于另一个点时，
    # 此时加不加1都可，加上也行，此时np.searchsorted函数的特性就是加上对应的超过的个数值，是我们所允许发生的
    # <class 'tuple'>: (150,)
    neg_inds = raw_samp + np.searchsorted(pos_inds_adj, raw_samp, side='right')

    return neg_inds


# train_matrix: <class 'tuple'>: (17052, 35119)
def generate_negative_samples(train_matrix, num_neg=3, num_sets=10):
    # list(17052)   neg_samples[0]: <class 'tuple'>: (150,)
    neg_samples = []
    for user_id, row in enumerate(train_matrix):
        # 可参考右文链接中的属性indices的意义，很清晰，存储了这些非零元素对应的列下标，即该user关联的item：https://zhuanlan.zhihu.com/p/108134307
        pos_ind = row.indices
        # <class 'tuple'>: (150,) 可见是真毒地每个user都逐一产生自己的负样本
        neg_sample = negsamp_vectorized_bsearch_preverif(pos_ind, train_matrix.shape[1], num_neg * num_sets)
        neg_samples.append(neg_sample)

    # <class 'tuple'>: (50, 17052, 3)
    final = np.asarray(neg_samples).reshape(num_sets, train_matrix.shape[0], num_neg)
    return final


def train_model(train_data, test_data, args):
    best_recall_epoch_num = 0
    best_ndcg_epoch_num = 0
    best_recall = float('-inf')
    best_ndcg = float('-inf')
    early_stop_acc = 0

    num_users = train_data.num_users
    num_items = train_data.num_items

    # convert to sequences, targets and users
    # <class 'tuple'>: (251785, 5)
    sequences_np = train_data.sequences.sequences
    targets_np = train_data.sequences.targets
    users_np = train_data.sequences.user_ids
    # <class 'tuple'>: (17052, 35119)
    train_matrix = train_data.tocsr()

    # 251785
    n_train = sequences_np.shape[0]
    print("Total training records:{}".format(n_train))

    model = HGN(num_users, num_items, args, device).to(device)

    if args.has_pretrain:
        if os.path.isfile(f'best_check/{args.data}_best.pt'):
            best_check = torch.load(f'best_check/{args.data}_best.pt')
            # 预训练
            model.user_embeddings.weight.data = best_check['user_embeddings.weight']
            model.item_embeddings.weight.data = best_check['item_embeddings.weight']
            model.W2.weight.data = best_check['W2.weight']
            model.b2.weight.data = best_check['b2.weight']
        else:
            print(f'pretrain file {args.data}_best.pt does not exist')

    # nelement：Returns the total number of elements in the input tensor，可参看官网，解释的很清楚
    print(f'The model has {count_parameters(model):,} trainable parameters')
    print('start training...')

    optimizer = torch.optim.Adam(model.parameters(), lr=args.learning_rate, weight_decay=args.l2)

    record_indexes = np.arange(n_train)
    # 4096
    batch_size = args.batch_size
    # 62
    num_batches = int(n_train / batch_size) + 1
    eval_epoch = 1 if args.data in ['ml', 'child', 'comics'] else 20
    for epoch_num in range(args.epochs):

        t1 = time()

        # set model to training mode
        model.train()

        np.random.shuffle(record_indexes)

        t_neg_start = time()
        # <class 'tuple'>: (50, 17052, 3)
        negatives_np_multi = generate_negative_samples(train_matrix, args.neg_samples, args.sets_of_neg_samples)
        # logger.info("Negative sampling time: {}s".format(time() - t_neg_start))

        epoch_loss = 0.0
        for batchID in range(num_batches):
            start = batchID * batch_size
            end = start + batch_size

            if batchID == num_batches - 1:
                if start < n_train:
                    end = n_train
                else:
                    break

            batch_record_index = record_indexes[start:end]

            # torch.Size([4096])
            batch_users = users_np[batch_record_index]
            # torch.Size([4096, 5])
            batch_sequences = sequences_np[batch_record_index]
            # torch.Size([4096, 3])
            batch_targets = targets_np[batch_record_index]
            # 这种情况看来作者是想每过一定的batch数后更换一批负样本
            # <class 'tuple'>: (17052, 3)
            negatives_np = negatives_np_multi[batchID % args.sets_of_neg_samples]
            # <class 'tuple'>: (4096, 3)
            batch_neg = negatives_np[batch_users]

            batch_users = torch.from_numpy(batch_users).type(torch.LongTensor).to(device)
            batch_sequences = torch.from_numpy(batch_sequences).type(torch.LongTensor).to(device)
            batch_targets = torch.from_numpy(batch_targets).type(torch.LongTensor).to(device)
            batch_negatives = torch.from_numpy(batch_neg).type(torch.LongTensor).to(device)

            # torch.Size([4096, 6])
            items_to_predict = torch.cat((batch_targets, batch_negatives), 1)
            # torch.Size([4096, 6])
            prediction_score = model(batch_sequences, batch_users, items_to_predict, False)

            # torch.Size([4096, 3])    torch.Size([4096, 3])
            (targets_prediction, negatives_prediction) = torch.split(
                prediction_score, [batch_targets.size(1), batch_negatives.size(1)], dim=1)

            # compute the BPR loss
            # torch.Size([4096, 3])
            loss = -torch.log(torch.sigmoid(targets_prediction - negatives_prediction) + 1e-8)
            # temp = torch.sum(loss)  # torch.Size([])
            # torch.mean在这没有用，属于冗余
            # torch.Size([])
            loss = torch.mean(torch.sum(loss))

            epoch_loss += loss.item()

            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
        epoch_loss /= num_batches

        t2 = time()

        logger(f"Epoch: {epoch_num + 1}, time: {t2 - t1:.1f}s,  loss: {epoch_loss:.4f}", loss_file)
        print(f'Epoch: {epoch_num + 1}, time: {t2 - t1:.1f}s,  loss: {epoch_loss:.4f}')

        # 为了调试改成1了，待会记得改回20：if (epoch_num + 1) % 20 == 0:
        # 后面的“or (epoch_num + 1) == 1”是为了提前固定显存，避免后面才暴涨导致的显存不够而停止
        if (epoch_num + 1) % eval_epoch == 0 or (epoch_num + 1) == 1:
            model.eval()
            # 论文中topk选的是10
            recall, ndcg = evaluation(model, train_data, test_data, args, topk=20)
            # pre_msg = ' '.join(str(e) for e in precision)
            recall_msg = ' '.join(str(e) for e in recall)
            # map_msg = ' '.join(str(e) for e in MAP)
            ndcg_msg = ' '.join(str(e) for e in ndcg)
            # logger(f'precision: {pre_msg}', log_file)
            logger(f'recall: {recall_msg}', log_file)
            print(f'recall: {recall_msg}')
            # logger(f'MAP: {map_msg}', log_file)
            logger(f'ndcg: {ndcg_msg}', log_file)
            print(f'ndcg: {ndcg_msg}')
            # logger("Evaluation time: {time() - t2}", log_file)

            # round(recall[0], 4)是为了小数点后数字过多导致的微小误差
            if recall[0] > best_recall or (round(recall[0], 4) == round(best_recall, 4) and ndcg[0] > best_ndcg):
                best_recall, best_ndcg, best_recall_epoch_num = recall[0], ndcg[0], epoch_num + 1
                early_stop_acc = 0
            else:
                # 早停
                early_stop_acc = early_stop_acc + 1
                if early_stop_acc > 10:
                    break

            # if recall[0] > best_recall or (recall[0] == best_recall and ndcg[0] > best_ndcg):
            #     best_recall = recall[0]
            #     best_recall_epoch_num = epoch_num + 1
            # if ndcg[0] > best_ndcg:
            #     best_ndcg = ndcg[0]
            #     best_ndcg_epoch_num = epoch_num + 1
    logger(f'best epoch number: {best_recall_epoch_num}, best recall: {best_recall}, best ndcg: {best_ndcg}', log_file)
    print(f'best epoch number: {best_recall_epoch_num}, best recall: {best_recall}, best ndcg: {best_ndcg}')
    logger_pure("\n\n", log_file)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()

    # data arguments
    parser.add_argument('--L', type=int, default=5)
    parser.add_argument('--T', type=int, default=3)

    # train arguments
    parser.add_argument('--data', type=str, default='cds')
    parser.add_argument('--epochs', type=int, default=200)
    parser.add_argument('--seed', type=int, default=1234)
    # 为了调试batch改小了，记得调回去4096
    parser.add_argument('--batch_size', type=int, default=4096)
    parser.add_argument('--batch_eval', type=int, default=512)
    parser.add_argument('--learning_rate', type=float, default=1e-3)
    parser.add_argument('--l2', type=float, default=1e-3)
    parser.add_argument('--neg_samples', type=int, default=3)
    # 作者是想每过一定的batch数后更换一批负样本
    parser.add_argument('--sets_of_neg_samples', type=int, default=50)

    # model dependent arguments
    parser.add_argument('--debug', type=bool, default=True)
    parser.add_argument('--d_model', type=int, default=50)
    parser.add_argument('--conv_layer', type=int, default=2)
    parser.add_argument('--kernel_size', type=int, default=3)

    parser.add_argument('--log_dir', type=str, default='log_dir', help='log_dir')

    args = parser.parse_args()
    args.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    if args.data in ['books', 'cds']:
        args.has_drop = True
        args.has_pretrain = True
    else:
        args.has_drop = False
        args.has_pretrain = True
    print(f'has dropout: {args.has_drop}')

    # 获取log文件名称，因为不同数据集实验起的名称不一样，同一个数据集每次也要单独分配一个名称记录
    log_file = get_curr_log_name(args)
    loss_file = get_curr_loss_name(args)
    # 记录下超参数, 加载checkpoint时是不用再记录一次的
    log_the_parameters(args, log_file)
    # 加上下面这行是为了避免创建log文件之后到创建loss文件之前程序意外终止导致的后续的log和loss文件编号的不一致
    logger('', loss_file)

    logger(datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"), log_file)
    print(f'{datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")}')

    if args.data == 'ml':
        data_set = MovieLens.ML20M()
    elif args.data == 'child':
        data_set = GoodReads.Children()
    elif args.data == 'comics':
        data_set = GoodReads.Comics()
    elif args.data == 'books':
        data_set = Amazon.Books()
    elif args.data == 'cds':
        data_set = Amazon.CDs()
    else:
        raise Exception("No suitable dataset found ...")

    # data_set = Amazon.CDs()  # MovieLens.ML20M()  # Books, CDs, Electronics

    # item_id=0 for sequence padding
    # train_set: list(17052)  train_set[1]:<class 'list'>: [16608, 101, 23667, 9860, 6569, 31417, 15923, 31906]
    # val_set: list(17052)  val_set[0]:<class 'list'>: [5118, 4695, 8869]  val_set[1]: <class 'list'>: [21971]
    # train_val_set: list(17052)  train_val_set[1]: <class 'list'>: [16608, 101, 23667, 9860, 6569, 31417, 15923, 31906, 21971]
    # test_set: list(17052)  test_set[0]: <class 'list'>: [10831, 17933, 15225, 18094, 34885, 26395, 15489]
    # num_users: 17052   num_items: 35119
    train_set, val_set, train_val_set, test_set, num_users, num_items = data_set.generate_dataset(index_shift=1)
    train = Interactions(train_val_set, num_users, num_items)
    train.to_sequence(args.L, args.T)

    train_model(train, test_set, args)
